from .fields import ConnectionFieldCachingMixin, DjangoConnectionField
from .optimizer import optimize
from .types import DjangoObjectType

__all__ = [
    "ConnectionFieldCachingMixin",
    "DjangoConnectionField",
    "DjangoObjectType",
    "optimize",
]



import graphene_django
from django.db.models import Model, QuerySet
from graphene.relay.connection import Connection
from graphene.types.definitions import GrapheneObjectType
from graphql.type.definition import GraphQLNonNull

from .optimizer import optimize
from .typing import PK, PK_CACHE_KEY, GQLInfo, TypeVar, Union

TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "DjangoObjectType",
]


class DjangoObjectType(graphene_django.types.DjangoObjectType):
    class Meta:
        abstract = True

    @classmethod
    def can_optimize_resolver(cls, info: GQLInfo) -> bool:
        return_type = info.return_type
        if isinstance(return_type, GraphQLNonNull):
            return_type = return_type.of_type

        return isinstance(return_type, GrapheneObjectType) and (
            issubclass(return_type.graphene_type, (cls, Connection))
        )

    @classmethod
    def get_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        if cls.can_optimize_resolver(info):
            queryset = optimize(queryset, info)
        return queryset

    @classmethod
    def get_node(cls, info: GQLInfo, id: PK) -> Union[TModel, None]:  # noqa: A002
        queryset: QuerySet[TModel] = cls._meta.model.objects.all()
        setattr(queryset, PK_CACHE_KEY, id)
        queryset = cls.get_queryset(queryset, info)
        return queryset.first()



import graphene_django.filter
from django.db.models import Model, QuerySet
from django.db.models.manager import Manager
from graphene.relay.connection import Connection
from graphql_relay.connection.connection import ConnectionType

from .cache import store_in_query_cache
from .optimizer import QueryOptimizer
from .typing import Any, Callable, GQLInfo, TypeVar, Union
from .utils import get_field_type, get_selections

TModel = TypeVar("TModel", bound=Model)

__all__ = [
    "ConnectionFieldCachingMixin",
    "DjangoConnectionField",
]


class ConnectionFieldCachingMixin:
    @classmethod
    def connection_resolver(  # noqa: PLR0913
        cls,
        resolver: Callable[..., Union[Manager[TModel], None]],
        connection: Connection,
        default_manager: Manager[TModel],
        queryset_resolver: Callable[..., QuerySet[TModel]],
        max_limit: int,
        enforce_first_or_last: bool,  # noqa: FBT001
        root: Union[Model, None],
        info: GQLInfo,
        **args: Any,
    ) -> ConnectionType:
        connection_field: graphene_django.fields.DjangoConnectionField = super()  # type: ignore[override]
        connection_type: ConnectionType = connection_field.connection_resolver(
            resolver,
            connection,
            default_manager,
            queryset_resolver,
            max_limit,
            enforce_first_or_last,
            root,
            info,
            **args,
        )
        if not connection_type.edges:
            return connection_type

        field_type = get_field_type(info)
        selections = get_selections(info)
        optimizer = QueryOptimizer(info)
        store = optimizer.optimize_selections(field_type, selections)
        store_in_query_cache(
            key=info.operation,
            items=(edge.node for edge in connection_type.edges),
            schema=info.schema,
            store=store,
        )

        return connection_type


class DjangoConnectionField(
    ConnectionFieldCachingMixin,
    graphene_django.fields.DjangoConnectionField,
):
    pass



from django.db.models import ManyToOneRel, Model, QuerySet
from graphene.relay.connection import ConnectionOptions
from graphene.types.definitions import GrapheneObjectType
from graphene.utils.str_converters import to_snake_case
from graphene_django.types import DjangoObjectTypeOptions
from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLField,
    InlineFragmentNode,
    SelectionNode,
)

from .cache import get_from_query_cache, store_in_query_cache
from .store import QueryOptimizerStore
from .typing import PK, PK_CACHE_KEY, GQLInfo, ModelField, ToManyField, TypeVar, Union
from .utils import get_field_type, get_selections, get_underlying_type, is_foreign_key_id, is_to_many, is_to_one

TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "optimize",
    "QueryOptimizer",
]


def optimize(queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
    field_type = get_field_type(info)
    selections = get_selections(info)
    if not selections:  # pragma: no cover
        return queryset

    optimizer = QueryOptimizer(info)
    store = optimizer.optimize_selections(field_type, selections)

    # Pk stored in 'query_optimizer.types.DjangoObjectType.get_node'
    pk: PK = getattr(queryset, PK_CACHE_KEY, None)

    if pk is not None:
        cached_item = get_from_query_cache(info.operation, info.schema, queryset.model, pk, store)
        if cached_item is not None:
            queryset._result_cache = [cached_item]
            return queryset

    queryset = store.optimize_queryset(queryset, pk=pk)
    if optimizer.cache_results:
        store_in_query_cache(info.operation, queryset, info.schema, store)

    return queryset


class QueryOptimizer:
    def __init__(self, info: GQLInfo) -> None:
        self.info = info
        self.cache_results = True

    def optimize_selections(
        self,
        field_type: GrapheneObjectType,
        selections: tuple[SelectionNode, ...],
    ) -> QueryOptimizerStore:
        store = QueryOptimizerStore()

        for selection in selections:
            if isinstance(selection, FieldNode):
                self.optimize_field_node(field_type, selection, store)

            elif isinstance(selection, FragmentSpreadNode):  # pragma: no cover
                self.optimize_fragment_spread(field_type, selection, store)

            elif isinstance(selection, InlineFragmentNode):  # pragma: no cover
                self.optimize_inline_fragment(selection)

            else:  # pragma: no cover
                msg = f"Unhandled selection node: '{selection}'"
                raise TypeError(msg)

        return store

    def optimize_field_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        options: Union[DjangoObjectTypeOptions, ConnectionOptions] = field_type.graphene_type._meta

        if isinstance(options, ConnectionOptions):
            return self.handle_relay_node(field_type, selection, store)

        elif isinstance(options, DjangoObjectTypeOptions):
            return self.handle_regular_node(field_type, selection, store)

        else:  # pragma: no cover
            msg = f"Unhandled field options type: {options}"
            raise TypeError(msg)

    # noinspection PyTypeChecker
    def handle_regular_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        model: type[Model] = field_type.graphene_type._meta.model
        selection_graphql_name = selection.name.value
        selection_graphql_field = field_type.fields.get(selection_graphql_name)
        if selection_graphql_field is None:  # pragma: no cover
            return

        model_field_name = to_snake_case(selection_graphql_name)
        model_field: ModelField = model._meta.get_field(model_field_name)

        if not model_field.is_relation or is_foreign_key_id(model_field, model_field_name):
            store.only(model_field_name)

        elif is_to_one(model_field):
            self.handle_to_one(model_field_name, selection, selection_graphql_field, store)

        elif is_to_many(model_field):
            self.handle_to_many(model_field_name, selection, selection_graphql_field, model_field, store)

        else:  # pragma: no cover
            msg = f">>> Unhandled selection: '{selection.name.value}'"
            raise ValueError(msg)

    def handle_relay_node(
        self,
        field_type: GrapheneObjectType,
        selection: FieldNode,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        # Connection QuerySets are sliced, so the results
        # should be cached later in the connection field.
        self.cache_results = False

        node: FieldNode = selection.selection_set.selections[0]  # type: ignore[assignment]
        if node.selection_set is None:  # page info
            return

        edges_field = field_type.fields[selection.name.value]
        edge_type = get_underlying_type(edges_field.type)
        node_field = edge_type.fields[node.name.value]
        node_type = get_underlying_type(node_field.type)

        nested_store = self.optimize_selections(node_type, node.selection_set.selections)
        store += nested_store

    def handle_to_one(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_graphql_field: GraphQLField,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        selection_field_type = get_underlying_type(selection_graphql_field.type)
        nested_store = self.optimize_selections(
            selection_field_type,
            selection.selection_set.selections,
        )

        store.select_related(model_field_name, nested_store)

    def handle_to_many(
        self,
        model_field_name: str,
        selection: FieldNode,
        selection_graphql_field: GraphQLField,
        model_field: ToManyField,
        store: QueryOptimizerStore,
    ) -> None:
        if selection.selection_set is None:  # pragma: no cover
            return

        selection_field_type = get_underlying_type(selection_graphql_field.type)
        nested_store = self.optimize_selections(
            selection_field_type,
            selection.selection_set.selections,
        )

        if isinstance(model_field, ManyToOneRel):  # Add connecting ID
            nested_store.only(model_field.field.name)

        related_queryset: QuerySet[Model] = model_field.related_model.objects.all()
        store.prefetch_related(model_field_name, nested_store, related_queryset)

    def optimize_fragment_spread(
        self,
        field_type: GrapheneObjectType,
        selection: FragmentSpreadNode,
        store: QueryOptimizerStore,
    ) -> None:  # pragma: no cover
        graphql_name = selection.name.value
        field_node = self.info.fragments[graphql_name]
        selections = field_node.selection_set.selections
        fragment_store = self.optimize_selections(field_type, selections)
        store.only_fields += fragment_store.only_fields
        store.select_stores.update(fragment_store.select_stores)
        store.prefetch_stores.update(fragment_store.prefetch_stores)

    def optimize_inline_fragment(self, selection: InlineFragmentNode) -> None:  # pragma: no cover
        fragment_type_name = selection.type_condition.name.value
        msg = f"TODO: {fragment_type_name}"
        raise NotImplementedError(msg)



from django.db.models import Model, Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP

from .typing import PK, TypeVar

TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "QueryOptimizerStore",
]


class QueryOptimizerStore:
    def __init__(self) -> None:
        self.only_fields: list[str] = []
        self.select_stores: dict[str, "QueryOptimizerStore"] = {}
        self.prefetch_stores: dict[str, tuple["QueryOptimizerStore", QuerySet[Model]]] = {}

    def only(self, field: str) -> None:
        self.only_fields.append(field)

    def select_related(self, name: str, store: "QueryOptimizerStore") -> None:
        self.select_stores[name] = store

    def prefetch_related(self, name: str, store: "QueryOptimizerStore", queryset: QuerySet[Model]) -> None:
        self.prefetch_stores[name] = (store, queryset)

    def compile(self, *, in_prefetch: bool = False) -> tuple[list[str], list[str], list[Prefetch]]:  # noqa: A003
        only_fields: list[str] = self.only_fields.copy()
        select_related: list[str] = []
        prefetch_related: list[Prefetch] = []

        for name, store in self.select_stores.items():
            if not in_prefetch:
                select_related.append(name)
            self._compile_nested(store, name, only_fields, select_related, prefetch_related, in_prefetch=in_prefetch)

        for name, (store, queryset) in self.prefetch_stores.items():
            optimized_queryset = store.optimize_queryset(queryset)
            prefetch_related.append(Prefetch(name, optimized_queryset))
            self._compile_nested(store, name, only_fields, select_related, prefetch_related, in_prefetch=True)

        return only_fields, select_related, prefetch_related

    @staticmethod
    def _compile_nested(
        store: "QueryOptimizerStore",
        name: str,
        only_fields: list[str],
        select_related: list[str],
        prefetch_related: list[Prefetch],
        *,
        in_prefetch: bool,
    ) -> None:
        store_only_list, store_select_related, store_prefetch_related = store.compile(in_prefetch=in_prefetch)
        if not in_prefetch:
            only_fields.extend(name + LOOKUP_SEP + only for only in store_only_list)
        select_related.extend(name + LOOKUP_SEP + select for select in store_select_related)
        for prefetch in store_prefetch_related:
            prefetch.add_prefix(name)
            prefetch_related.append(prefetch)

    def optimize_queryset(self, queryset: QuerySet[TModel], *, pk: PK = None) -> QuerySet[TModel]:
        only_fields, select_related, prefetch_related = self.compile()

        if only_fields:
            queryset = queryset.only(*only_fields)
        if select_related:
            queryset = queryset.select_related(*select_related)
        if prefetch_related:
            queryset = queryset.prefetch_related(*prefetch_related)
        if pk is not None:
            queryset = queryset.filter(pk=pk)

        return queryset

    def __add__(self, other: "QueryOptimizerStore") -> "QueryOptimizerStore":
        self.only_fields += other.only_fields
        self.select_stores.update(other.select_stores)
        self.prefetch_stores.update(other.prefetch_stores)
        return self

    def __str__(self) -> str:
        only_fields, select_related, prefetch_related = self.compile()
        only = ",".join(only_fields)
        select = ",".join(select_related)
        prefetch = ",".join(item.prefetch_to for item in prefetch_related)
        return f"{only=}|{select=}|{prefetch=}"



from collections import defaultdict
from typing import TYPE_CHECKING, Union
from weakref import WeakKeyDictionary

from django.db.models import Model, QuerySet
from graphql import GraphQLSchema

from .typing import PK, Hashable, Iterable, QueryCache, TableName, TypeVar

if TYPE_CHECKING:
    from .store import QueryOptimizerStore


TModel = TypeVar("TModel", bound=Model)


__all__ = [
    "get_from_query_cache",
    "get_query_cache",
    "store_in_query_cache",
]


def get_query_cache(key: Hashable, schema: GraphQLSchema) -> QueryCache:
    cache = schema.extensions.setdefault("_query_cache", WeakKeyDictionary())
    return cache.setdefault(key, defaultdict(lambda: defaultdict(dict)))  # type: ignore[no-any-return]


def get_from_query_cache(
    key: Hashable,
    schema: GraphQLSchema,
    model: type[TModel],
    pk: PK,
    store: "QueryOptimizerStore",
) -> Union[TModel, None]:
    """
    Get model instance from query cache for the given 'field_type'.
    Key should be any hashable value that is present only for the duration of
    a single request, e.g., 'info.operation'.
    """
    store_str = str(store)
    query_cache = get_query_cache(key, schema)
    return query_cache[model._meta.db_table][store_str].get(pk)


def store_in_query_cache(
    key: Hashable,
    items: Iterable[Model],
    schema: GraphQLSchema,
    store: "QueryOptimizerStore",
) -> None:
    """
    Set all models in list (as well as any select related models)
    to the query cache for the given 'field_type'.
    Key should be any hashable value that is present only for the duration of
    a single request, e.g., 'info.operation'.
    """
    query_cache = get_query_cache(key, schema)
    items = list(items)
    annotations = _get_annotations(items[0])
    for item in items:
        _add_item(query_cache, item, annotations, store)


def _get_annotations(item: Model) -> list[str]:
    # Don't use 'queryset.query.annotations' since we cannot extract annotations
    # as cleanly from the model if some other iterable is given.
    # (these results contain foreign key ids as well)
    model_builtins = {"_prefetched_objects_cache", "_state"}
    fields: set[str] = {field.name for field in item._meta.get_fields() if field.name not in model_builtins}
    attributes: set[str] = set(item.__dict__)
    return list(attributes.difference(fields))


def _add_item(query_cache: QueryCache, instance: Model, annotations: list[str], store: "QueryOptimizerStore") -> None:
    store_str = str(store)
    if annotations:
        store_str += f"|{annotations=}"
    table_name: TableName = instance._meta.db_table
    query_cache[table_name][store_str][instance.pk] = instance
    _add_selected(query_cache, instance, store)
    _add_prefetched(query_cache, instance, store)


def _add_selected(query_cache: QueryCache, instance: Model, store: "QueryOptimizerStore") -> None:
    for nested_name, nested_store in store.select_stores.items():
        nested_instance: Model = getattr(instance, nested_name)
        _add_item(query_cache, nested_instance, [], nested_store)


def _add_prefetched(query_cache: QueryCache, instance: Model, store: "QueryOptimizerStore") -> None:
    for nested_name, (nested_store, _queryset) in store.prefetch_stores.items():
        selected: QuerySet[Model] = getattr(instance, nested_name).all()
        for select in selected:
            _add_item(query_cache, select, [], nested_store)
